@@ -1,13 +1,11 @@
 #!/usr/bin/perl -w
-# 
-# $Id: colordiff.pl,v 1.4.2.12 2005/03/28 20:11:34 daveewart Exp $
 
 ########################################################################
 #                                                                      #
 # ColorDiff - a wrapper/replacment for 'diff' producing                #
 #             colourful output                                         #
 #                                                                      #
-# Copyright (C)2002-2004 Dave Ewart (davee@sungate.co.uk)              #
+# Copyright (C)2002-2006 Dave Ewart (davee@sungate.co.uk)              #
 #                                                                      #
 ########################################################################
 #                                                                      #
@@ -21,10 +19,6 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
 # GNU General Public License for more details.                         #
 #                                                                      #
-# You should have received a copy of the GNU General Public License    #
-# along with this program; if not, write to the Free Software          #
-# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            #
-#                                                                      #
 ########################################################################
 
 use strict;
@@ -32,12 +26,13 @@ use Getopt::Long qw(:config pass_through);
 use IPC::Open2;
 
 my $app_name     = 'colordiff';
-my $version      = '1.0.5';
+my $version      = '1.0.6';
 my $author       = 'Dave Ewart';
 my $author_email = 'davee@sungate.co.uk';
 my $app_www      = 'http://colordiff.sourceforge.net/';
-my $copyright    = '(C)2002-2005';
+my $copyright    = '(C)2002-2006';
 my $show_banner  = 1;
+my $color_patch  = 0;
 
 # ANSI sequences for colours
 my %colour;
@@ -57,7 +52,7 @@ $colour{darkcyan}    = "\033[0;36m";
 $colour{darkred}     = "\033[0;31m";
 $colour{darkmagenta} = "\033[0;35m";
 $colour{darkblack}   = "\033[0;30m";
-$colour{OFF}         = "\033[0;0m";
+$colour{off}         = "\033[0;0m";
 
 # Default colours if /etc/colordiffrc or ~/.colordiffrc do not exist
 my $plain_text = $colour{white};
@@ -91,8 +86,17 @@ foreach $config_file (@config_files) {
                 }
                 next;
             }
+            if ($setting eq 'color_patch') {
+                if ($value eq 'yes') {
+                    $color_patch = 1;
+                }
+                next;
+            }
             $setting =~ tr/A-Z/a-z/;
             $value =~ tr/A-Z/a-z/;
+            if (($value eq 'normal') || ($value eq 'none')) {
+                $value = 'off';
+            }
             if (!defined $colour{$value}) {
                 print STDERR "Invalid colour specification ($value) in $config_file\n";
                 next;
@@ -120,7 +124,18 @@ foreach $config_file (@config_files) {
     }
 }
 
-# colordiff specfic options here.  Need to pre-declare if using variables
+# If output is to a file, switch off colours, unless 'color_patch' is set
+# Relates to http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=378563
+if ((-f STDOUT) && ($color_patch == 0)) {
+    $plain_text='';
+    $file_old='';
+    $file_new='';
+    $diff_stuff='';
+    $cvs_stuff='';
+    $plain_text='';
+    $colour{off}='';
+}
+
 
 # ----------------------------------------------------------------------------
 # FIXME
@@ -167,6 +182,7 @@ else {
 
 my $diff_type = 'unknown';
 my $record;
+my $longest_record = 0;
 
 DIFF_TYPE: foreach $record (@inputstream) {
     # Unified diffs are the only flavour having '+++' or '---'
@@ -186,10 +202,82 @@ DIFF_TYPE: foreach $record (@inputstream) {
         $diff_type = 'diff';
         last DIFF_TYPE;
     }
+    # FIXME - This is not very specific, since the regex matches could
+    # easily match non-diff output.
+    # However, given that we have not yet matched any of the *other* diff
+    # types, this might be good enough
+    elsif ($record =~ /(\s\|\s|\s<$|\s>\s)/) {
+        $diff_type = 'diffy';
+        last DIFF_TYPE;
+    }
 }
 
 my $inside_file_old = 1;
 
+# ------------------------------------------------------------------------------
+# Special pre-processing for side-by-side diffs
+# Figure out location of central markers: these will be a consecutive set of
+# three columns where the first and third always consist of spaces and the
+# second consists only of spaces, '<', '>' and '|'
+# This is not a 100% certain match, but should be good enough
+
+my %separator_col = ();
+my %candidate_col = ();
+my $diffy_sep_col = 0;
+my $mostlikely_sum = 0;
+
+if ($diff_type eq 'diffy') {
+    # Not very elegant, but does the job
+    # Unfortunately requires parsing the input stream multiple times
+    foreach (@inputstream) {
+        # Convert tabs to spaces
+        while ((my $i = index($_, "\t")) > -1) {
+            substr(
+                    $_, $i, 1,                # range to replace
+                    (' ' x (8 - ($i % 8))),     # string to replace with
+                  );
+        }
+        $record = $_;
+        $longest_record = length($record) if (length($record) > $longest_record);
+    }
+    for(my $i=0;$i <= $longest_record;$i++) {
+        $separator_col{$i} = 1;
+        $candidate_col{$i} = 0;
+    }
+
+    foreach (@inputstream) {
+        # Convert tabs to spaces
+        while ((my $i = index($_, "\t")) > -1) {
+            substr(
+                    $_, $i, 1,                # range to replace
+                    (' ' x (8 - ($i % 8))),     # string to replace with
+                  );
+        }
+        for(my $i=0;$i<(length($_)-2);$i++) {
+            next if (!defined $separator_col{$i});
+            next if ($separator_col{$i} == 0);
+            my $subsub = substr($_,$i,2);
+            if ( ($subsub ne '  ') && ($subsub ne ' |') && ($subsub ne ' >') &&
+                    ($subsub ne ' <') ) {
+                $separator_col{$i} = 0;
+            }
+            if (($subsub eq ' |') || ($subsub eq ' >') || ($subsub eq ' <')) {
+                $candidate_col{$i}++;
+            }
+        }
+    }
+
+    for(my $i=0;$i<$longest_record-2;$i++) {
+        if ($separator_col{$i} == 1) {
+            if ($candidate_col{$i} > $mostlikely_sum) {
+                $diffy_sep_col = $i;
+                $mostlikely_sum = $i;
+            }
+        }
+    }
+}
+# ------------------------------------------------------------------------------
+
 foreach (@inputstream) {
     if ($diff_type eq 'diff') {
         if (/^</) {
@@ -264,13 +352,36 @@ foreach (@inputstream) {
             print "$plain_text";
         }
     }
-    s/$/$colour{OFF}/;
+    # Works with previously-identified column containing the diff-y
+    # separator characters
+    elsif ($diff_type eq 'diffy') {
+        if (length($_) > ($diffy_sep_col+2)) {
+            my $sepchars = substr($_, $diffy_sep_col, 2);
+            if ($sepchars eq ' <') {
+                print "$file_old";
+            }
+            elsif ($sepchars eq ' |') {
+                print "$diff_stuff";
+            }
+            elsif ($sepchars eq ' >') {
+                print "$file_new";
+            }
+            else {
+                print "$plain_text";
+            }
+        }
+        else {
+            print "$plain_text";
+        }
+    }
+    s/$/$colour{off}/;
     print "$_";
 }
 
-sub set_color {
-    my ($type, $color) = @_;
-
-    $type =~ s/-/_/;
-    eval "\$$type = \$colour{$color}";
-}
+#   FIXME - See earlier commented-out code
+### sub set_color {
+###     my ($type, $color) = @_;
+### 
+###     $type =~ s/-/_/;
+###     eval "\$$type = \$colour{$color}";
+### }
